// Swarm library. Copyright  1999-2000 Swarm Development Group.
// This library is distributed without any warranty; without even the
// implied warranty of merchantability or fitness for a particular purpose.
// See file LICENSE for details and terms of copying.

package swarm;

import swarm.random.*;
import swarm.simtoolsgui.*;
import swarm.defobj.*;
import swarm.activity.*;
import swarm.objectbase.*;

public class SwarmEnvironment {
  public Symbol Start, Member, End;
  public Symbol Initialized, Running, Stopped, Holding, Released, 
      Terminated, Completed;
  public Symbol Randomized, Sequential;
  public Symbol ControlStateRunning, ControlStateStopped,
    ControlStateStepping, ControlStateQuit, ControlStateNextTime;
  public Zone scratchZone, globalZone;
  public SimpleRandomGenerator randomGenerator;
  public UniformIntegerDist uniformIntRand;
  public UniformDoubleDist uniformDblRand;
  public ProbeLibrary probeLibrary;
  public ProbeDisplayManager probeDisplayManager;
  public HDF5Archiver hdf5Archiver;
  public LispArchiver lispArchiver;
  public HDF5Archiver hdf5AppArchiver;
  public LispArchiver lispAppArchiver;
  public boolean guiFlag;
  public ClassLoader proxyClassLoader;

  public SwarmEnvironment () {
    super ();

    proxyClassLoader = new ProxyClassLoader ();

    globalZone = new ZoneImpl ();
    scratchZone = new ZoneImpl ();

    randomGenerator = new MT19937genImpl ();
    uniformIntRand = new UniformIntegerDistImpl ();
    uniformDblRand = new UniformDoubleDistImpl ();
    probeLibrary = new ProbeLibraryImpl ();
    probeDisplayManager = new ProbeDisplayManagerImpl ();

    ControlStateRunning = new SymbolImpl ();
    ControlStateStopped = new SymbolImpl ();
    ControlStateStepping = new SymbolImpl ();
    ControlStateQuit = new SymbolImpl ();
    ControlStateNextTime = new SymbolImpl ();

    Start = new SymbolImpl ();
    Member = new SymbolImpl ();
    End = new SymbolImpl ();

    Initialized = new SymbolImpl ();
    Running = new SymbolImpl ();
    Stopped = new SymbolImpl ();
    Holding = new SymbolImpl ();
    Released = new SymbolImpl ();
    Terminated = new SymbolImpl ();
    Completed = new SymbolImpl ();

    Randomized = new SymbolImpl ();
    Sequential = new SymbolImpl ();
    hdf5Archiver = new HDF5ArchiverImpl ();
    lispArchiver = new LispArchiverImpl ();
    hdf5AppArchiver = new HDF5ArchiverImpl ();
    lispAppArchiver = new LispArchiverImpl ();
  }
  public native void initSwarm (String appName, String version, String bugAddress, String args[]);
  public native void xprint (Object obj);
  public native void xfprint (Object obj);
  public native void dumpDirectory ();
  public native int getCurrentTime ();
  public native Swarm getCurrentSwarm ();
  public native Schedule getCurrentSchedule ();
  public native SwarmActivity getCurrentSwarmActivity ();
  public native ScheduleActivity getCurrentScheduleActivity();
  public native Activity getCurrentOwnerActivity ();
  public native Action getCurrentAction ();
  public native Activity getCurrentActivity ();
  
  public native Object createProbeDisplay (Object anOjbect);
  public native Object createCompleteProbeDisplay (Object anObject);
  public native Object createArchivedProbeDisplay (Object anObject, String objectName);
  public native Object createArchivedCompleteProbeDisplay (Object anObject, String objectName);
  public native Object setWindowGeometryRecordName (Object theWidget, String widgetName);
  public native Object setComponentWindowGeometryRecordNameFor (Object obj, Object widget, String widgetName);
  public native Object setComponentWindowGeometryRecordName (Object widget, String widgetName);
}


